# tests/test_point.py
"""
This module tests the Point class
"""
import unittest

from datetime import date, datetime
import json
import numpy as np

from thymus.timeseries import Timeseries
from thymus.point import Point


class TestPoint(unittest.TestCase):
    """ This class tests the class Point. """

    def setUp(self):

        # three timeseries
        self.ts = Timeseries()
        self.ts.key = "Test Key"
        self.ts.columns = ["dog", "cat", "squirrel"]

        self.start_date = datetime(2021, 1, 29).toordinal()
        self.ts.dseries = self.start_date + np.arange(5)
        self.ts.tseries = np.arange(15).reshape((5, 3)) / 10.33
        self.ts.make_arrays()

    def test_class_init_(self):
        """Test class initialization."""
        point = Point(self.ts, 3)

        self.assertTrue(hasattr(point, "dog"))
        self.assertTrue(hasattr(point, "cat"))
        self.assertTrue(hasattr(point, "squirrel"))

        self.assertEqual(point.dog, point.values[0])
        self.assertEqual(point.cat, point.values[1])
        self.assertEqual(point.squirrel, point.values[2])
        self.assertEqual(point.date, point.ts.dseries[3])
        self.assertEqual(point.row_no, 3)

        point.dog = 1
        point.cat = 2
        point.squirrel = 3

        self.assertEqual(point.dog, point.values[0])
        self.assertEqual(point.cat, point.values[1])
        self.assertEqual(point.squirrel, point.values[2])

    def test__repr__(self):
        """ Test the appearance. """
        point = Point(self.ts, 3)

        output = repr(point)

        self.assertTrue(output.startswith("<Point"))
        self.assertTrue(output.endswith("/>"))

        for column in self.ts.columns:
            with self.subTest(column=column):
                self.assertTrue(output.find(column) > -1)

        # no columns
        self.ts.columns = None
        point = Point(self.ts, 3)

        output = repr(point)

        # has values
        self.assertTrue(output.find("[") > -1)

        # should show new name
        class NewPoint(Point):
            def __init__(self, ts, row_no):
                super().__init__(ts, row_no)

        output = repr(NewPoint(self.ts, 3))

        self.assertTrue(output.startswith("<NewPoint"))
        self.assertTrue(output.endswith("/>"))

        # format question
        ts = self.ts[:, 1]
        point = Point(ts, 3)

        self.assertEqual(
            repr(point),
            "<Point: row_no: 3, date: 2021-02-01, 0.968054211035818 />",
        )

        ts.columns = ["test"]
        point = Point(ts, 3)

        self.assertEqual(
            repr(point),
            "<Point: row_no: 3, date: 2021-02-01, test: 0.968054211035818 />",
        )

    def test_to_dict(self):
        """ Test formatting for a dictionary. """

        # native date format
        self.assertDictEqual(
            Point(self.ts, 3).to_dict(),
            {
                "row_no": 3,
                "date": datetime(2021, 1, 29).toordinal() + 3,
                "dog": 0.8712487899322362,
                "cat": 0.968054211035818,
                "squirrel": 1.0648596321393997,
            },
        )

        # str date format
        self.assertDictEqual(
            Point(self.ts, 0).to_dict(dt_fmt="str"),
            {
                "row_no": 0,
                "date": datetime(2021, 1, 29).strftime("%F"),
                "dog": 0.0,
                "cat": 0.0968054211035818,
                "squirrel": 0.1936108422071636,
            },
        )

        # str date format
        self.assertDictEqual(
            Point(self.ts, 0).to_dict(dt_fmt="datetime"),
            {
                "row_no": 0,
                "date": datetime(2021, 1, 29).date(),
                "dog": 0.0,
                "cat": 0.0968054211035818,
                "squirrel": 0.1936108422071636,
            },
        )

        # invalid date format
        self.assertRaises(ValueError, Point(self.ts, 0).to_dict, dt_fmt="test")


if __name__ == "__main__":
    unittest.main()



# thymus/point.py
"""
This module implements the Point class.
"""


class Point(object):
    """
    This class enables a combined look at a specific point in a
    time series. While the purpose of Thymus-Timeseries is the
    separation of dates and values, there are times when having
    both together can be very useful.

    Usage:
    point = Point(ts, row)

    This produces an object:

    point.ts      a reference to the parent time series
    point.date    ordinal/timestamp
    point.values  an array of values

    If columns are being used for the object, properties
    are created for each column / value pair.

    For example:

    ts.columns = ["dog", "cat", "squirrel"]}

    Properties will have
        point.dog
        point.cat
        point.squirrel

    as well as values, as referenced above.

    Setting point.dog to a new value will update the column in
    ts.tseries[row].

    print(point)
    <Point: row_no: 3, date: 2020-01-04,
      dog: 0.8709958385754379
      cat: 0.9677731539727088
      squirrel: 1.0645504693699797 />

    """

    def __init__(self, ts, row_no):
        self.ts = ts
        self.row_no = row_no

        if len(self.ts.tseries.shape) == 2:
            shape = self.ts.tseries.shape
        else:
            shape = (self.ts.tseries.shape, 1)

        if ts.columns and len(ts.columns) == shape[1]:

            for idx, column in enumerate(ts.columns):

                def get_value(self, idx=idx):
                    return self.values[idx]

                get_value.__doc__ = f"Gets value from column {idx}."

                def set_value(self, value, idx=idx):
                    self.values[idx] = value

                set_value.__doc__ = f"Sets value in column {idx}."

                setattr(self.__class__, column, property(get_value, set_value))

    @property
    def values(self):
        """
        The values found at ts.tseries[row_no].

        These values cannot be changed directly. However, that is
        possible using the generated column variables.
        """
        return self.ts.tseries[self.row_no]

    @property
    def date(self):
        """ The date found at ts.dseries[row_no]. """
        return self.ts.dseries[self.row_no]

    def __repr__(self, line_break=60):
        """
        If the line of column/values will be over 60 characters, the
        output will be vertical for easier viewing.
        """
        if len(self.ts.tseries.shape) == 1:
            # wrong shape
            if self.ts.columns:
                values = f"{self.ts.columns[0]}: {self.values}"
            else:
                values = self.values

        elif self.ts.columns:
            values = ", ".join(
                [
                    f"{column}: {value}"
                    for column, value in zip(self.ts.columns, self.values)
                ]
            )
            if len(values) > line_break:
                values = "\n  " + values.replace(", ", "\n  ")
        else:
            values = self.values

        sdate = self.date_str()
        name = self.__class__.__name__
        row_no = self.row_no
        return f"<{name}: row_no: {row_no}, date: {sdate}, {values} />"

    def date_str(self):
        """
        Returns the date in string format.
        """
        return self.ts.fmt_date(self.date, self.ts.get_date_series_type())

    def datetime(self):
        """
        Returns the date as a date/datetime object.
        """
        return self.ts.get_datetime(self.date)

    def to_dict(self, dt_fmt=None):
        """
        This function returns a dict of the point variables.

        Usage:
            to_dict(dt_fmt=None)

        Parameters:
            dt_fmt: (None|str) : Format choice is "str" or "datetime"

        Returns:
            point (dict)
        """

        pdict = {"row_no": self.row_no}

        if dt_fmt is None:
            pdict["date"] = self.date
        elif dt_fmt == "str":
            pdict["date"] = self.date_str()
        elif dt_fmt == "datetime":
            pdict["date"] = self.datetime()
        else:
            raise ValueError("Invalid dt_fmt: choices(None,str, datetime)")
        if self.ts.columns:
            tmp = dict(
                [
                    [column, value]
                    for column, value in zip(self.ts.columns, self.values)
                ]
            )

            pdict.update(tmp)
        return pdict
